//! Generate register interface rust code from types in [`crate::types`].

use crate::types::{Block, Module, ModuleBlockElements};
use proc_macro2::TokenStream;
use quote::quote;

pub trait CodeGen {
    fn generate_register_interface(self) -> proc_macro2::TokenStream;
}

impl CodeGen for Module {
    fn generate_register_interface(self) -> proc_macro2::TokenStream {
        let mut out = TokenStream::new();
        if !self.elements_bitstring.is_empty() {
            todo!("bitstring generation is not yet implemented")
        }

        let child_mods = self
            .elements_other
            .into_iter()
            .map(|e| e.generate_register_interface());
        for child in child_mods {
            out.extend(child);
        }

        out
    }
}

impl CodeGen for ModuleBlockElements {
    fn generate_register_interface(self) -> proc_macro2::TokenStream {
        match self {
            ModuleBlockElements::Block(block) => block.generate_register_interface(),
            ModuleBlockElements::Register(register) => todo!(),
            ModuleBlockElements::Memory(memory) => todo!(),
            ModuleBlockElements::Fifo(fifo) => todo!(),
        }
    }
}

impl CodeGen for Block {
    fn generate_register_interface(self) -> proc_macro2::TokenStream {
        let name = self.name;
        let desc = self.desc.unwrap_or("".to_string());

        quote! {
            pub mod #name {
                #[doc = #desc]
                
                use std::marker::PhantomData;
                
                use super::RegisterInterface;
                
                pub mod control;
                pub mod status;
                
                pub(crate) const SLR0_OFFSET: usize = 0x0000;
                pub(crate) const SLR0_SIZE: usize = 0x4000;
                
                /// Access to SLR0.
                pub struct Slr0<'a> {
                    mem_ptr: *mut u32,
                    _marker: PhantomData<&'a mut RegisterInterface>,
                }
                
                impl Slr0<'_> {
                    pub(crate) fn new(ptr: *mut u32) -> Self {
                        Slr0 {
                            mem_ptr: ptr,
                            _marker: PhantomData,
                        }
                    }
                    pub fn control(&self) -> control::Control {
                        control::Control::new(unsafe { self.mem_ptr.add(control::OFFSET) })
                    }
                
                    pub fn status(&self) -> status::Status {
                        status::Status::new(unsafe { self.mem_ptr.add(status::OFFSET) })
                    }
                }
            }
        }
    }
}
